#!/usr/bin/env python
# coding=utf-8

import logging
import math
import threading
import time
import os, sys
from bosftp_config import BosftpCfg
from multipart_upload_helper import MultipartUploadHelper
from thread_pool import BosThreadPool
from client import Client
from utils import FifoBuffer
from baidubce import exception
from baidubce.services import bos
logger = logging.getLogger(__name__)

class Uploader(object):
    """upload parts in multi thread """
    def __init__(self, bucket_name, object_key):
        self.name = object_key
        self.init_already = False
        self.uploader = MultipartUploadHelper(bucket_name, object_key)
        self.failed_parts = []
        self.succeed_parts = []
        self.condition = threading.Condition()
        self.part_size = BosftpCfg().multi_part_size
        self.closed = False
        self.part_num = 1
        self.parts_remain = 0
        self.buffer = FifoBuffer() 
    # file write interface
    def write(self, data):
        logger.info("Recv data size: %d, file: %s" % (len(data), self.uploader.object_key))         
        # put 0 size object
        if len(data) == 0:
            Client().bos_client.put_object_from_string(self.uploader.bucket_name, self.uploader.object_key, "")
            
        if len(self.failed_parts) > 0:
            raise IOError("some parts upload failed")
        # accumulate data to multi_part_size
        self.buffer.write(data)
    
        # send data asynchronously
        while self.buffer.size() >= self.part_size:
            self._async_upload_data(self.uploader.bucket_name, 
                                    self.uploader.object_key, self.buffer.read(self.part_size))
        
        return len(data)
    # file close interface
    def close(self):
        logger.info("Closing file: {0}".format(self.uploader.object_key))
        if self.uploader.upload_id == None:
            logger.info("upload_id is None, call put_object_from_file, filename: {0}".format(self.filename))
            Client().bos_client.put_object_from_string(self.uploader.bucket_name, 
                                                       self.uploader.object_key, 
                                                       self.buffer.read(self.buffer.size()))
        # if last part has not been sent, send it
        if self.buffer.size() > 0:
            logger.info(
                "send last part of file: {0}, last part size: {1}, part_num: {2}".format(
                    self.uploader.object_key, self.buffer.size(), self.part_num))
            self._async_upload_data(
                self.uploader.bucket_name, self.uploader.object_key, self.buffer.read(self.buffer.size()))
        if self.init_already:
            # complete_upload
            self.condition.acquire()
            while len(self.failed_parts) == 0 and self.parts_remain != 0:
                self.condition.wait()
            self.condition.release()
            # some part(s) failed
            if len(self.failed_parts) != 0:
                logger.error("upload part failed, part_num: %d" % int(self.failed_parts[0]))
                raise IOError("upload part failed, part_num: %d" % int(self.failed_parts[0]))
            # all parts upload success, now complete them 
            logger.debug("all parts are uploaded success, now we complete them")
            try:
                self.uploader.complete_upload()    
            except exception.BceError as e:
                raise IOError("complete_upload failed, upload_id: %d" % int(self.uploader.upload_id))
        self.buffer.close()
        self.closed = True
          
    # upload data(self.part_size normally multi_part_size) asynchronously 
    def _async_upload_data(self, bucket_name, object_key, data):
        if data == None:
            logger.info("data is None, return")
            return
        
        # init
        if not self.init_already:
            try:
                self.uploader.init()
                self.init_already = True
            except exception.BceError as e:
                logger.error("init uploader failed, bucket_name: {0}, object_key: {1}, err_msg: {2}".format(str(self.bucket_name).encode("utf-8"), str(self.object_key).encode("utf-8"), e.message))
                raise e
        logger.debug(
                "parallel_upload, upload_id: {0}, size: {1}, part_size: {2}".format(
                self.uploader.upload_id, len(data), self.part_size))
        # callback for notify from blocking 
        def callback(part_num, res):
            logger.debug("callback, part_num:{0}, res:{1}".format(part_num, res))
            self.condition.acquire()
            if not res:
                logger.error("upload_part occurs an error, part_num: {0}".format(part_num))
                self.failed_parts.append(part_num)
                self.condition.notify()
            else:
                self.parts_remain = self.parts_remain - 1
                logger.debug("callback, self.parts_remain: %d" % (self.parts_remain))
                if self.parts_remain == 0:
                    self.condition.notify()
            self.condition.release()
        # multi thread upload
        self.condition.acquire()
        self.parts_remain = self.parts_remain + 1
        self.condition.release()
        logger.debug("BosThreadPool().apply_async, part_size: %d, part_num: %d" % (int(self.part_size), self.part_num))
        BosThreadPool().apply_async(self.uploader.upload_part,
                    (self.part_num, data, callback))
        self.part_num = self.part_num + 1
